open Common
open Poly

type t = Leq of term * term
       | Eq of term * term
       | Not of t
       | Or of t * t
       | And of t * t
       | Forall of (var -> t)

let implies p q = Or(Not p, q)

open Core
module E = Z3.Expr
module A = Z3.Arithmetic
module R = Z3.Arithmetic.Real

module M = struct
  include Multi

  let rec to_Z3 ctx m = function
  | Const c     -> R.mk_numeral_s ctx (Q.to_string c)
  | Var x       -> begin match Hashtbl.find m x with
    | Some x -> x
    | None   -> let s = E.mk_const_s ctx x (R.mk_sort ctx) in begin
        ignore (Hashtbl.add m ~key:x ~data:s);
        s
      end
  end
  | Plus(l, r)  -> A.mk_add ctx [to_Z3 ctx m l; to_Z3 ctx m r]
  | Times(l, r) -> A.mk_mul ctx [to_Z3 ctx m l; to_Z3 ctx m r]
end

let rec subst x e = function
| Leq(l, r) -> Leq(M.subst x e l, M.subst x e r)
| Eq(l, r)  -> Eq(M.subst x e l, M.subst x e r)
| Not t     -> Not(subst x e t)
| Or(l, r)  -> Or(subst x e l, subst x e r)
| And(l, r) -> And(subst x e l, subst x e r)
(* Weak HOAS encoding gives capture-avoiding substitution for free,
   as long as y is instantiated with a variable not in the free variables of the body.
   we handle that by disallowing user variables from starting with underscores, so we
   can use them. *)
| Forall t  -> Forall(fun y -> subst x e (t y))

let count = ref 0

let fresh () =
  let n = !count in
  count := n + 1;
  "_x" ^ string_of_int n

let descope () =
  count := !count - 1

module B = Z3.Boolean
module Q = Z3.Quantifier

let rec to_Z3 ctx m = function
| Leq(l, r)      -> A.mk_le ctx (M.to_Z3 ctx m l) (M.to_Z3 ctx m r)
| Not(Leq(l, r)) -> A.mk_gt ctx (M.to_Z3 ctx m l) (M.to_Z3 ctx m r)
| Eq(l, r)       -> B.mk_eq ctx (M.to_Z3 ctx m l) (M.to_Z3 ctx m r)
| Not t          -> B.mk_not ctx (to_Z3 ctx m t)
| Or(Not l, r)   -> B.mk_implies ctx (to_Z3 ctx m l) (to_Z3 ctx m r)
| Or(l, r)       -> B.mk_or ctx [to_Z3 ctx m l; to_Z3 ctx m r]
| And(l, r)      -> B.mk_and ctx [to_Z3 ctx m l; to_Z3 ctx m r]
| Forall t  ->
    let x = fresh () in
    let f = Q.expr_of_quantifier
      (Q.mk_forall_const ctx [R.mk_const_s ctx x] (to_Z3 ctx m (t x)) None [] [] None None) in
    descope ();
    f

module S = Z3.Solver
module T = Z3.Tactic
module G = Z3.Goal

(* http://blog.emmatosch.com/2017/03/27/calling-external-programs.html *)
let exec_program content =
  let (fname, out) = Filename.open_temp_file "_constraints" "txt" in
  Out_channel.output_string out content;
  Out_channel.flush out;
  let stdout = Unix.open_process_in ("z3 -smt2 " ^ fname) in
  In_channel.input_all stdout

let check t =
  (* Translate FOL formula to Z3 constraint *)
  Z3.set_global_param "smt.pull_nested_quantifiers" "true";
  let ctx    = Z3.mk_context ["model", "true"; "proof", "true"] in
  let m      = Hashtbl.create (module String) in
  let f      = to_Z3 ctx m (Not t) in
  count     := 0;
  let frees = Hashtbl.fold m
    ~init:"(set-option :smt.pull_nested_quantifiers true)\n(set-option :produce-proofs true)\n"
    ~f:(fun ~key ~data s -> ignore data;
         if String.get key 0 = '_' then s
         else s ^ Printf.sprintf "(declare-const %s Real)\n" key) in
  let program = Printf.sprintf "%s\n(assert %s)\n(check-sat)" frees (E.to_string f) in
  print_endline ("Querying Z3 with:\n\n" ^ program);
  (* Query Z3 if Not(t) is satisfiable. If it is, then t is invalid, else it is. *)
  let result = exec_program program in
  if String.is_substring result ~substring:"unsat" then
    Valid (exec_program (program ^ "\n(get-proof)"))
  else if String.is_substring result ~substring:"sat" then
    Invalid (exec_program (program ^ "\n(get-model)"))
  else
    Unknown result
